/*
##############################################################################
#
# Copyright by The HDF Group.
# All rights reserved.
#
# This file is part of the hl_region High-Level HDF5 APIs.  The full copyright
# notice, including terms governing use, modification, and redistribution,
# is contained in the file COPYING, which can be found at the root of the
# source code distribution tree and in the documentation directory (doc/html/).
# If you do not have access to this file, you may request a copy of
# "the hl_region High-Level HDF5 APIs copyright and license statement" from
# help@hdfgroup.org.
#
##############################################################################
*/

/*
 *****************************************************************************
 * PURPOSE
 *   This file contains C stubs for hl_region_H5LR Fortran APIs
 ******
*/
#include "hdf5.h"
#include "H5f90i_gen.h"
#include "H5f90i.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include "hl_regionf90proto.h"
#include "../../src/h5hl_region.h"

/*-------------------------------------------------------------------------
 * Function: H5LRget_region_info_c
 *
 * Purpose: Calls H5LRget_region_info
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: January 20, 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRget_region_info_c(hid_t_f *obj_id, void *ref, size_t_f *pathlen,                
		      _fcd path, int_f *path_size, int_f *rank, hid_t_f *dtype, 
		      int_f *sel_type, size_t_f *numelem, void *buf)
{
  herr_t status;
  int ret_value = -1;

  char *c_path = NULL;
  size_t c_pathlen;
  int c_rank = 1;
  int c_sel_type = 1;
  hid_t c_dtype = 1;
  size_t c_numelem = 1;
  hsize_t *c_block_coord;
  int i,ii,j;

  /*
   * Allocate buffer to hold name of an object
   */
  c_pathlen = 0;

  if(*pathlen != 0) {
    c_path = (char *)malloc((size_t)*pathlen+1);
    c_pathlen = (size_t)*pathlen+1;
    if (c_path == NULL) return ret_value;
  }
  if( *rank < 0 ) c_rank = 0;
  if( *dtype < 0 ) c_dtype = 0;
  c_sel_type = 0;
  if( *numelem == 0) { 
    c_numelem = 0;
  }
  else { 
    if( *numelem > 0 ) c_numelem = (size_t)*numelem;
  }

  /*
   * Call H5LRget_region_info
   */
  status = H5LRget_region_info((hid_t)*obj_id, ref, &c_pathlen,
			       c_path, &c_rank, &c_dtype, &c_sel_type,
			       &c_numelem, buf);

  if(*path_size > 0) {
    /*
     * Convert C name to FORTRAN and place it in the given buffer
     */
    HD5packFstring(c_path, _fcdtocp(path), (size_t)*path_size);
  }
 
  /* subtract from the length the NULL termination character. */
  if(c_pathlen !=0 )
    *pathlen = (size_t_f)c_pathlen-1;

  *rank = (int_f)c_rank;
  *dtype = (hid_t_f)c_dtype;
  *sel_type = (int_f)c_sel_type;
  *numelem = (hid_t_f)c_numelem;

  if ( status >= 0  ) ret_value = 0;

  if(c_path)
    free(c_path);

  return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: H5LRcreate_region_references_c
 *
 * Purpose: Calls H5LRcreate_region_references
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: January 20, 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRcreate_region_references_c(hid_t_f *file_id, size_t_f *num_elem, char *path, 
			       int_f *char_len, int_f *max_char_size, 
			       int_f *block_coord_size, 
			       const hsize_t_f *block_coord, void *ref_buf )
{
  herr_t status;
  int ret_value = -1;

  char **c_path = NULL;
  size_t *c_pathlen;
  hsize_t *c_block_coord;
  int i,ii,j, k, kk;
  hid_t dset_id = -1;      /* destination: dataset */
  hid_t space = -1;        /* identifier of the dataspace */
  int nrank;               /* the dimensionality of the source dataspace */
  int nrank2;              /* the dimensionality of the source dataspace */
  int *c_ranks;
  char *tmp_p;

  /*
   * Because of the difference in C-FORTRAN storage order the rank of each region has to be
   * determined so that the hyperslab descriptions contained in block_coord can be 
   * transposed correctly. Additionaly, subtract 1 to account for C starting arrays at 0.
   */

  /*
   * Allocate array of character pointers
   */
  c_path = (char **)malloc((size_t)*num_elem * sizeof(char *));
  if (c_path == NULL) return ret_value;
  
  /* Copy data to long C string */
  tmp_p = (char *)HD5f2cstring(path, (size_t)*max_char_size*(size_t)*num_elem);
  if (tmp_p == NULL) { free(c_path);
    return ret_value;
  }

  /*
   * Move data from temporary buffer
   */
  for (i=0; i < *num_elem; i++) {
    c_path[i] = (char *) malloc((size_t)char_len[i]+1);
    memcpy(c_path[i], tmp_p, (size_t)char_len[i]);
    c_path[i][char_len[i]] = '\0';
    tmp_p = tmp_p + *max_char_size;
  }
  c_ranks=(int *)malloc(sizeof(int)*(*num_elem));
  if(!c_ranks) return ret_value;
  nrank2 = 0;
  for(i=0; i< *num_elem; i++) {

    /* Open the dataset for a given the path */
    dset_id = H5Dopen2((hid_t)*file_id, c_path[i], H5P_DEFAULT);
    if(dset_id < 0)
      return ret_value;

    /* Get the dataspace of the dataset */
    space = H5Dget_space(dset_id);

    if(space < 0)
      return ret_value;

    /* Find the rank of the dataspace */
    nrank = H5Sget_simple_extent_ndims(space);

    if(nrank < 0)
      return ret_value;
    c_ranks[i] = nrank;
    nrank2 = nrank2 + nrank*2;

  }

  c_block_coord = (hsize_t *)malloc(sizeof(hsize_t)*nrank2);
  if(!c_block_coord) return ret_value;

  /*
   * Transpose dimension arrays because of C-FORTRAN storage order
   * and subtract 1 to account for C starting arrays at 0. Additionally,
   * take into account the fact that the rank of each data set 
   * may be different.
   */
  kk = 0;
  for(k=0; k< *num_elem; k++){
    ii = 0;
    for (j = 0; j < 2; j++) {
      for (i = 0; i < c_ranks[k]; i++) {
	c_block_coord[ii+kk] = block_coord[(j+1)*(c_ranks[k])-i-1+kk]-1;
	ii = ii + 1;
      }
    }
    kk = kk + 2*c_ranks[k];
  }

  /*
   * Call H5LRcreate_region_references
   */
  status = H5LRcreate_region_references( (hid_t)*file_id, (size_t)*num_elem, (const char **)c_path, c_block_coord, ref_buf ); 

  if ( status >= 0  ) ret_value = 0;

  for (i=0; i < *num_elem; i++) {
     if(c_path[i])
         free(c_path[i]);
  }
  if(c_path)
      free(c_path);
  
  free(c_block_coord);
  
  free(c_ranks);
  
  return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: H5LRread_region_c 
 *
 * Purpose: Calls H5LRread_region
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: February, 6 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRread_region_c(hid_t_f *loc_id, void *ref, hid_t_f *mem_type, size_t_f *num_elem, void *buf )
{
  herr_t status;
  int ret_value = -1;

  size_t c_num_elem;

  status = H5LRread_region( (hid_t)*loc_id, ref, (hid_t)*mem_type, &c_num_elem, buf);

  *num_elem = (size_t_f)c_num_elem;
 
  if ( status >= 0  ) ret_value = 0;

  return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: H5LRmake_dataset_c
 *
 * Purpose: Calls H5LRmake_dataset
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: February 10, 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRmake_dataset_c(hid_t_f *loc_id, size_t_f *pathlen,                
		      char *path, hid_t_f *type_id, size_t_f *buf_size,
		      void *loc_id_ref, void *ref)
{
  herr_t status;
  int ret_value = -1;

  char *c_path  = NULL;

  /*
   * Convert FORTRAN name to C name
   */
  if(NULL == (c_path = (char *)HD5f2cstring(path, (size_t)*pathlen)))
    return ret_value;

   /*
    * Call H5LRmake_dataset
    */
  status = H5LRmake_dataset_FORTRAN((hid_t)*loc_id, c_path,
			    (hid_t)*type_id,(size_t)*buf_size,
			    loc_id_ref, ref);
  if(c_path)
    free(c_path);

  if ( status >= 0  ) ret_value = 0;

  return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: H5LRcopy_region_c
 *
 * Purpose: Calls H5LRcopy_region
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: February 20, 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRcopy_region_c(hid_t_f *obj_id, void *ref, size_t_f *file_dest_len,                
		  char *file_dest, size_t_f *path_dest_len, char *path_dest, int_f *rank,
		  hsize_t_f *block_coord_dest)
{

  herr_t status;
  int ret_value = -1;

  char *c_file_dest  = NULL;
  char *c_path_dest  = NULL;
  hsize_t *c_block_coord;
  int i, ii, j;

  /*
   * Convert FORTRAN name to C name
   */
  if(NULL == (c_file_dest = (char *)HD5f2cstring(file_dest, (size_t)*file_dest_len)))
    return ret_value;
  if(NULL == (c_path_dest = (char *)HD5f2cstring(path_dest, (size_t)*path_dest_len)))
    return ret_value;

  c_block_coord = (hsize_t *)malloc(sizeof(hsize_t)*(*rank*2));
  if(!c_block_coord) return ret_value;
  /*
   * Transpose dimension arrays because of C-FORTRAN storage order
   * and subtract 1 to account for C starting arrays at 0.
   */
   ii = 0;
   for (j = 0; j < 2; j++) {
     for (i = 0; i < *rank ; i++) {
       c_block_coord[ii] =  block_coord_dest[(j+1)*(*rank) - i -1] - 1;
       ii = ii + 1;
     }
   }

   /*
    * Call H5LRcopy_region
    */

  status = H5LRcopy_region((hid_t)*obj_id, ref, c_file_dest, c_path_dest, c_block_coord);

  if ( status >= 0  ) ret_value = 0;

  if(c_path_dest)
    free(c_path_dest);
  if(c_file_dest)
    free(c_file_dest);

  free(c_block_coord);

  return ret_value;
}
/*-------------------------------------------------------------------------
 * Function: H5LRcopy_reference_c
 *
 * Purpose: Calls H5LRcopy_reference
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: March 1, 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRcopy_reference_c(hid_t_f *obj_id, void *ref, size_t_f *file_len,                
		     char *file, size_t_f *path_len, char *path, int_f *rank,
		     hsize_t_f *block_coord, void *ref_new)
{
  herr_t status;
  int ret_value = -1;
  
  char *c_file  = NULL;
  char *c_path  = NULL;
  hsize_t *c_block_coord;
  int i, ii, j;

  /*
   * Convert FORTRAN name to C name
   */
  if(NULL == (c_file = (char *)HD5f2cstring(file, (size_t)*file_len)))
    return ret_value;
  if(NULL == (c_path = (char *)HD5f2cstring(path, (size_t)*path_len)))
    return ret_value;
  
  c_block_coord = (hsize_t *)malloc(sizeof(hsize_t)*(*rank*2));
  if(!c_block_coord) return ret_value;
  /*
   * Transpose dimension arrays because of C-FORTRAN storage order
   * and subtract 1 to account for C starting arrays at 0.
   */
  ii = 0;
  for (j = 0; j < 2; j++) {
    for (i = 0; i < *rank ; i++) {
      c_block_coord[ii] =  block_coord[(j+1)*(*rank) - i -1] - 1;
      ii = ii + 1;
    }
  }
  /*
   * Call H5LRcopy_reference
   */
  status = H5LRcopy_reference((hid_t)*obj_id, ref, c_file, c_path, c_block_coord, ref_new);

  if ( status >= 0  ) ret_value = 0;

  if(c_path)
    free(c_path);
  if(c_file)
    free(c_file);

  free(c_block_coord);

  return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: H5LRcreate_ref_to_all_c
 *
 * Purpose: Calls H5LRcreate_ref_to_all
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: March 1, 2010
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LRcreate_ref_to_all_c(hid_t_f *loc_id, size_t_f *group_path_len, char *group_path, 
			size_t_f *ds_path_len, char *ds_path, 
			int_f *index_type, int_f *order, int_f *ref_type)
{
  
  herr_t status;
  int ret_value = -1;
  
  char *c_group_path = NULL;
  char *c_ds_path = NULL;

  /*
   * Convert FORTRAN name to C name
   */
  if(NULL == (c_group_path = (char *)HD5f2cstring(group_path, (size_t)*group_path_len)))
    return ret_value;
  if(NULL == (c_ds_path = (char *)HD5f2cstring(ds_path, (size_t)*ds_path_len)))
    return ret_value;
  
  /*
   * Call H5LRcreate_ref_to_all
   */
  status = H5LRcreate_ref_to_all((hid_t)*loc_id, c_group_path, c_ds_path, 
				 (H5_index_t)*index_type, (H5_iter_order_t)*order, (H5R_type_t)*ref_type);

  if ( status >= 0  ) ret_value = 0;

  if(c_group_path)
    free(c_group_path);
  if(c_ds_path)
    free(c_ds_path);

  return ret_value;
}
